@@ -17,15 +17,20 @@
 #include <QDBusConnectionInterface>
 #include <QDBusReply>
 
-#include "KDBusServiceIface.h"
 #include "FreeDesktopApplpicationIface.h"
+#include "KDBusServiceIface.h"
 
 #include "config-kdbusaddons.h"
 
 #if HAVE_X11
+#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+#include <private/qtx11extras_p.h>
+#else
 #include <QX11Info>
 #endif
+#endif
 
+#include "kdbusaddons_debug.h"
 #include "kdbusservice_adaptor.h"
 #include "kdbusserviceextensions_adaptor.h"
 
@@ -33,19 +38,16 @@ class KDBusServicePrivate
 {
 public:
     KDBusServicePrivate()
-        : registered(false),
-          exitValue(0)
-    {}
+        : registered(false)
+        , exitValue(0)
+    {
+    }
 
     QString generateServiceName()
     {
         const QCoreApplication *app = QCoreApplication::instance();
         const QString domain = app->organizationDomain();
-#if QT_VERSION < QT_VERSION_CHECK(5, 15, 0)
-        const QStringList parts = domain.split(QLatin1Char('.'), QString::SkipEmptyParts);
-#else
         const QStringList parts = domain.split(QLatin1Char('.'), Qt::SkipEmptyParts);
-#endif
 
         QString reversedDomain;
         if (parts.isEmpty()) {
@@ -60,6 +62,23 @@ public:
         return reversedDomain + app->applicationName();
     }
 
+    static void handlePlatformData(const QVariantMap &platformData)
+    {
+        #if HAVE_X11
+        if (QX11Info::isPlatformX11()) {
+            QByteArray desktopStartupId = platformData.value(QStringLiteral("desktop-startup-id")).toByteArray();
+            if (!desktopStartupId.isEmpty()) {
+                QX11Info::setNextStartupId(desktopStartupId);
+            }
+        }
+        #endif
+
+        const auto xdgActivationToken = platformData.value(QLatin1String("activation-token")).toByteArray();
+        if (!xdgActivationToken.isEmpty()) {
+            qputenv("XDG_ACTIVATION_TOKEN", xdgActivationToken);
+        }
+    }
+
     bool registered;
     QString serviceName;
     QString errorMessage;
@@ -67,12 +86,13 @@ public:
 };
 
 // Wraps a serviceName registration.
-class Registration : public QObject {
+class Registration : public QObject
+{
     Q_OBJECT
 public:
     enum class Register {
         RegisterWitoutQueue,
-        RegisterWithQueue
+        RegisterWithQueue,
     };
 
     Registration(KDBusService *s_, KDBusServicePrivate *d_, KDBusService::StartupOptions options_)
@@ -81,27 +101,27 @@ public:
         , options(options_)
     {
         if (!QDBusConnection::sessionBus().isConnected() || !(bus = QDBusConnection::sessionBus().interface())) {
-            d->errorMessage = QLatin1String("Session bus not found\n"
-                                            "To circumvent this problem try the following command (with Linux and bash)\n"
-                                            "export $(dbus-launch)");
+            d->errorMessage = QLatin1String(
+                "DBus session bus not found. To circumvent this problem try the following command (with bash):\n"
+                "    export $(dbus-launch)");
         } else {
             generateServiceName();
         }
     }
 
-    void run() {
+    void run()
+    {
         if (bus) {
             registerOnBus();
         }
 
         if (!d->registered && ((options & KDBusService::NoExitOnFailure) == 0)) {
-            qCritical() << d->errorMessage;
+            qCCritical(KDBUSADDONS_LOG) << qPrintable(d->errorMessage);
             exit(1);
         }
     }
 
-    private:
-
+private:
     void generateServiceName()
     {
         d->serviceName = d->generateServiceName();
@@ -110,9 +130,10 @@ public:
         objectPath.replace(QLatin1Char('-'), QLatin1Char('_')); // see spec change at https://bugs.freedesktop.org/show_bug.cgi?id=95129
 
         if (options & KDBusService::Multiple) {
-            const bool inSandbox = QFileInfo::exists(QStringLiteral("/flatpak-info"));
+            const bool inSandbox = QFileInfo::exists(QStringLiteral("/.flatpak-info"));
             if (inSandbox) {
-                d->serviceName += QStringLiteral(".kdbus-") + QDBusConnection::sessionBus().baseService().replace(QRegularExpression(QStringLiteral("[\\.:]")), QStringLiteral("_"));
+                d->serviceName += QStringLiteral(".kdbus-")
+                    + QDBusConnection::sessionBus().baseService().replace(QRegularExpression(QStringLiteral("[\\.:]")), QStringLiteral("_"));
             } else {
                 d->serviceName += QLatin1Char('-') + QString::number(QCoreApplication::applicationPid());
             }
@@ -125,17 +146,17 @@ public:
         bool objectRegistered = false;
         objectRegistered = bus.registerObject(QStringLiteral("/MainApplication"),
                                               QCoreApplication::instance(),
-                                              QDBusConnection::ExportAllSlots |
-                                              QDBusConnection::ExportScriptableProperties |
-                                              QDBusConnection::ExportAdaptors);
+                                              QDBusConnection::ExportAllSlots //
+                                                  | QDBusConnection::ExportScriptableProperties //
+                                                  | QDBusConnection::ExportAdaptors);
         if (!objectRegistered) {
-            qWarning() << "Failed to register /MainApplication on DBus";
+            qCWarning(KDBUSADDONS_LOG) << "Failed to register /MainApplication on DBus";
             return;
         }
 
         objectRegistered = bus.registerObject(objectPath, s, QDBusConnection::ExportAdaptors);
         if (!objectRegistered) {
-            qWarning() << "Failed to register" << objectPath << "on DBus";
+            qCWarning(KDBUSADDONS_LOG) << "Failed to register" << objectPath << "on DBus";
             return;
         }
 
@@ -143,8 +164,7 @@ public:
 
         if (d->registered) {
             if (QCoreApplication *app = QCoreApplication::instance()) {
-                connect(app, &QCoreApplication::aboutToQuit,
-                        s, &KDBusService::unregister);
+                connect(app, &QCoreApplication::aboutToQuit, s, &KDBusService::unregister);
             }
         }
     }
@@ -165,8 +185,7 @@ public:
 
             queueOption = QDBusConnectionInterface::QueueService;
 
-            connect(bus, &QDBusConnectionInterface::serviceRegistered,
-                    this, [this](const QString &service) {
+            connect(bus, &QDBusConnectionInterface::serviceRegistered, this, [this](const QString &service) {
                 if (service != d->serviceName) {
                     return;
                 }
@@ -176,8 +195,7 @@ public:
             });
         }
 
-        d->registered =
-                (bus->registerService(d->serviceName, queueOption) == QDBusConnectionInterface::ServiceRegistered);
+        d->registered = (bus->registerService(d->serviceName, queueOption) == QDBusConnectionInterface::ServiceRegistered);
 
         if (d->registered) {
             return;
@@ -185,15 +203,30 @@ public:
 
         if (options & KDBusService::Replace) {
             auto message = QDBusMessage::createMethodCall(d->serviceName,
-                                                    QStringLiteral("/MainApplication"),
-                                                    QStringLiteral("org.qtproject.Qt.QCoreApplication"),
-                                                    QStringLiteral("quit"));
+                                                          QStringLiteral("/MainApplication"),
+                                                          QStringLiteral("org.qtproject.Qt.QCoreApplication"),
+                                                          QStringLiteral("quit"));
             QDBusConnection::sessionBus().asyncCall(message);
             waitForRegistration();
         } else if (options & KDBusService::Unique) {
             // Already running so it's ok!
             QVariantMap platform_data;
-            platform_data.insert(QStringLiteral("desktop-startup-id"), QString::fromUtf8(qgetenv("DESKTOP_STARTUP_ID")));
+#if HAVE_X11
+            if (QX11Info::isPlatformX11()) {
+                QString startupId = QString::fromUtf8(qgetenv("DESKTOP_STARTUP_ID"));
+                if (startupId.isEmpty()) {
+                    startupId = QString::fromUtf8(QX11Info::nextStartupId());
+                }
+                if (!startupId.isEmpty()) {
+                    platform_data.insert(QStringLiteral("desktop-startup-id"), startupId);
+                }
+            }
+#endif
+
+            if (qEnvironmentVariableIsSet("XDG_ACTIVATION_TOKEN")) {
+                platform_data.insert(QStringLiteral("activation-token"), qgetenv("XDG_ACTIVATION_TOKEN"));
+            }
+
             if (QCoreApplication::arguments().count() > 1) {
                 OrgKdeKDBusServiceInterface iface(d->serviceName, objectPath, QDBusConnection::sessionBus());
                 iface.setTimeout(5 * 60 * 1000); // Application can take time to answer
@@ -220,9 +253,7 @@ public:
         }
 
         if (!d->registered) { // either multi service or failed to reclaim name
-            d->errorMessage = QLatin1String("Couldn't register name '")
-                    + d->serviceName
-                    + QLatin1String("' with DBUS - another process owns it already!");
+            d->errorMessage = QLatin1String("Couldn't register name '") + d->serviceName + QLatin1String("' with DBUS - another process owns it already!");
         }
     }
 
@@ -246,19 +277,17 @@ public:
 };
 
 KDBusService::KDBusService(StartupOptions options, QObject *parent)
-    : QObject(parent), d(new KDBusServicePrivate)
+    : QObject(parent)
+    , d(new KDBusServicePrivate)
 {
     new KDBusServiceAdaptor(this);
     new KDBusServiceExtensionsAdaptor(this);
 
-    Registration registration(this, d, options);
+    Registration registration(this, d.get(), options);
     registration.run();
 }
 
-KDBusService::~KDBusService()
-{
-    delete d;
-}
+KDBusService::~KDBusService() = default;
 
 bool KDBusService::isRegistered() const
 {
@@ -291,50 +320,38 @@ void KDBusService::unregister()
 
 void KDBusService::Activate(const QVariantMap &platform_data)
 {
-    Q_UNUSED(platform_data);
-#if HAVE_X11
-    if (QX11Info::isPlatformX11()) {
-        QX11Info::setAppTime(QX11Info::getTimestamp());
-    }
-#endif
-    // TODO QX11Info::setNextStartupId(platform_data.value("desktop-startup-id"))
-    emit activateRequested(QStringList(), QString());
-    // TODO (via hook) KStartupInfo::appStarted(platform_data.value("desktop-startup-id"))
-    // ^^ same discussion as below
+    d->handlePlatformData(platform_data);
+    Q_EMIT activateRequested(QStringList(), QString());
+    qunsetenv("XDG_ACTIVATION_TOKEN");
 }
 
 void KDBusService::Open(const QStringList &uris, const QVariantMap &platform_data)
 {
-    Q_UNUSED(platform_data);
-    // TODO QX11Info::setNextStartupId(platform_data.value("desktop-startup-id"))
-    emit openRequested(QUrl::fromStringList(uris));
-    // TODO (via hook) KStartupInfo::appStarted(platform_data.value("desktop-startup-id"))
-    // ^^ not needed if the app actually opened a new window.
-    // Solution 1: do it all the time anyway (needs API in QX11Info)
-    // Solution 2: pass the id to the app and let it use KStartupInfo::appStarted if reusing a window
+    d->handlePlatformData(platform_data);
+    Q_EMIT openRequested(QUrl::fromStringList(uris));
+    qunsetenv("XDG_ACTIVATION_TOKEN");
 }
 
 void KDBusService::ActivateAction(const QString &action_name, const QVariantList &maybeParameter, const QVariantMap &platform_data)
 {
-    Q_UNUSED(platform_data);
+    d->handlePlatformData(platform_data);
+
     // This is a workaround for D-Bus not supporting null variants.
     const QVariant param = maybeParameter.count() == 1 ? maybeParameter.first() : QVariant();
-    emit activateActionRequested(action_name, param);
-    // TODO (via hook) KStartupInfo::appStarted(platform_data.value("desktop-startup-id"))
-    // if desktop-startup-id is set, the action is supposed to show a window (since it could be
-    // called when the app is not running)
+
+    Q_EMIT activateActionRequested(action_name, param);
+    qunsetenv("XDG_ACTIVATION_TOKEN");
 }
 
 int KDBusService::CommandLine(const QStringList &arguments, const QString &workingDirectory, const QVariantMap &platform_data)
 {
-    Q_UNUSED(platform_data);
     d->exitValue = 0;
+    d->handlePlatformData(platform_data);
     // The TODOs here only make sense if this method can be called from the GUI.
     // If it's for pure "usage in the terminal" then no startup notification got started.
     // But maybe one day the workspace wants to call this for the Exec key of a .desktop file?
-    // TODO QX11Info::setNextStartupId(platform_data.value("desktop-startup-id"))
-    emit activateRequested(arguments, workingDirectory);
-    // TODO (via hook) KStartupInfo::appStarted(platform_data.value("desktop-startup-id"))
+    Q_EMIT activateRequested(arguments, workingDirectory);
+    qunsetenv("XDG_ACTIVATION_TOKEN");
     return d->exitValue;
 }
 
